# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""RecordDraft Files Service API."""

from .config import FileMetadataServiceConfig, FileServiceConfig
from .service import FileMetadataService, FileService

__all__ = (
    'FileMetadataService',
    'FileMetadataServiceConfig',
    'FileService',
    'FileServiceConfig',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Errors."""

from flask_resources.errors import HTTPJSONException


class ActionNotImplementedError(HTTPJSONException):
    """Command not implemented error."""

    code = 500
    description = "Action not implemented."

    def __init__(self, cmd_name=None, **kwargs):
        """Constructor."""
        super(ActionNotImplementedError, self).__init__(**kwargs)
        if cmd_name:
            self.description = f"Action '{cmd_name}'' not implemented."



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio Drafts Resources module to create REST APIs."""

from flask import abort, g
from flask_resources import CollectionResource, SingletonResource
from flask_resources.context import resource_requestctx
from invenio_records_resources.config import ConfigLoaderMixin

from ...services import RecordDraftService
from .config import DraftActionResourceConfig, DraftResourceConfig, \
    DraftVersionResourceConfig
from .errors import ActionNotImplementedError


class DraftResource(SingletonResource, ConfigLoaderMixin):
    """Draft resource."""

    default_config = DraftResourceConfig

    def __init__(self, config=None, service=None):
        """Constructor."""
        super(DraftResource, self).__init__(config=self.load_config(config))
        self.service = service or RecordDraftService()

    def read(self):
        """Edit a draft.

        GET /records/:pid_value/draft
        """
        item = self.service.read_draft(
            resource_requestctx.route["pid_value"],
            g.identity,
            links_config=self.config.links_config,
        )
        return item.to_dict(), 200

    def create(self):
        """Edit a record.

        POST /records/:pid_value/draft
        """
        item = self.service.edit(
            resource_requestctx.route["pid_value"],
            g.identity,
            links_config=self.config.links_config,
        )
        return item.to_dict(), 201

    def update(self):
        """Update a draft.

        PUT /records/:pid_value/draft
        """
        data = resource_requestctx.request_content
        item = self.service.update_draft(
            resource_requestctx.route["pid_value"],
            g.identity,
            data,
            links_config=self.config.links_config,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return item.to_dict(), 200

    def delete(self):
        """Delete a draft.

        DELETE /records/:pid_value/draft
        """
        self.service.delete_draft(
            resource_requestctx.route["pid_value"],
            g.identity,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return None, 204


class DraftVersionResource(CollectionResource, ConfigLoaderMixin):
    """Draft version resource."""

    default_config = DraftVersionResourceConfig

    def __init__(self, service=None, config=None):
        """Constructor."""
        super(DraftVersionResource, self).__init__(
            config=self.load_config(config))
        self.service = service or RecordDraftService()

    def search(self):
        """Perform a search over the items.

        GET /records/:pid_value/versions
        """
        # TODO: IMPLEMENT ME!
        return self.service.search(), 200

    def create(self):
        """Create a new version.

        POST /records/:pid_value/versions
        """
        item = self.service.new_version(
            resource_requestctx.route["pid_value"],
            g.identity
        )
        return item.to_dict(), 201


class DraftActionResource(SingletonResource, ConfigLoaderMixin):
    """Draft action resource."""

    default_config = DraftActionResourceConfig

    def __init__(self, service=None, config=None):
        """Constructor."""
        super(DraftActionResource, self).__init__(
            config=self.load_config(config))
        self.service = service or RecordDraftService()

    def create(self):
        """Any POST business logic.

        POST /records/:pid_value/actions/:action
        """
        action = resource_requestctx.route["action"]
        try:
            cmd_name = self.config.action_commands[action]
            cmd_func = getattr(self.service, cmd_name)
        except KeyError:
            raise abort(404)
        except AttributeError:
            raise ActionNotImplementedError(cmd_name)

        if cmd_name == "publish":
            item = cmd_func(
                resource_requestctx.route["pid_value"],
                g.identity,
                links_config=self.config.record_links_config
            )
        else:
            item = cmd_func(
                resource_requestctx.route["pid_value"],
                g.identity,
            )

        return item.to_dict(), 202



"""Example of a record draft API."""

from invenio_records.systemfields import ConstantField
from invenio_records_resources.records.systemfields import IndexField

from invenio_drafts_resources.records.api import Draft as DraftBase
from invenio_drafts_resources.records.api import Record as RecordBase

from .models import DraftMetadata, RecordMetadata


class Record(RecordBase):
    """Example record API."""

    # Configuration
    model_cls = RecordMetadata

    # System fields
    schema = ConstantField(
        '$schema', 'http://localhost/schemas/records/record-v1.0.0.json')

    index = IndexField('records-record-v1.0.0', search_alias='records')


class Draft(DraftBase):
    """Example record API."""

    # Configuration
    model_cls = DraftMetadata

    # System fields
    schema = ConstantField(
        '$schema', 'http://localhost/schemas/records/record-v1.0.0.json')

    index = IndexField('drafts-draft-v1.0.0', search_alias='drafts')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio Drafts Resources module to create REST APIs"""

import json
import time

import pytest
from sqlalchemy.orm.exc import NoResultFound

from invenio_drafts_resources.resources import DraftActionResource, \
    DraftActionResourceConfig, DraftResource, DraftResourceConfig, \
    DraftVersionResource, DraftVersionResourceConfig

HEADERS = {"content-type": "application/json", "accept": "application/json"}


@pytest.fixture()
def input_data():
    """Input data (as coming from the view layer)."""
    return {
        'metadata': {
            'title': 'Test'
        },
    }


def _assert_single_item_response(response):
    """Assert the fields present on a single item response."""
    response_fields = response.json.keys()
    fields_to_check = ['id', 'conceptid', 'metadata',
                       'created', 'updated', 'links']

    for field in fields_to_check:
        assert field in response_fields

#
# Operations tests
#


def test_create_draft(client, input_data):
    """Test draft creation of a non-existing record."""
    response = client.post(
        "/mocks", data=json.dumps(input_data), headers=HEADERS)

    assert response.status_code == 201
    _assert_single_item_response(response)


def test_read_draft(client, input_data):
    response = client.post(
        "/mocks", data=json.dumps(input_data), headers=HEADERS)

    assert response.status_code == 201

    recid = response.json['id']
    response = client.get(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert response.status_code == 200

    _assert_single_item_response(response)


def test_update_draft(client, input_data):
    response = client.post(
        "/mocks", data=json.dumps(input_data), headers=HEADERS)

    assert response.status_code == 201
    assert response.json['metadata']['title'] == \
        input_data['metadata']['title']

    recid = response.json['id']

    orig_title = input_data['metadata']['title']
    edited_title = "Edited title"
    input_data['metadata']['title'] = edited_title

    # Update draft content
    update_response = client.put(
        "/mocks/{}/draft".format(recid),
        data=json.dumps(input_data),
        headers=HEADERS
    )

    assert update_response.status_code == 200
    assert update_response.json["metadata"]['title'] == edited_title
    assert update_response.json["id"] == recid

    # Check the updates where saved
    update_response = client.get(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert update_response.status_code == 200
    assert update_response.json["metadata"]['title'] == edited_title
    assert update_response.json["id"] == recid


def test_delete_draft(client, input_data):
    response = client.post(
        "/mocks", data=json.dumps(input_data), headers=HEADERS)

    assert response.status_code == 201

    recid = response.json['id']

    update_response = client.delete(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert update_response.status_code == 204

    # Check draft deletion
    # FIXME: Remove import when exception is properly handled
    with pytest.raises(NoResultFound):
        update_response = client.get(
            "/mocks/{}/draft".format(recid), headers=HEADERS)

        assert update_response.status_code == 404


def _create_and_publish(client, input_data):
    """Create a draft and publish it."""
    # Create the draft
    response = client.post(
        "/mocks", data=json.dumps(input_data), headers=HEADERS)

    assert response.status_code == 201
    recid = response.json['id']

    # Publish it
    response = client.post(
        "/mocks/{}/draft/actions/publish".format(recid), headers=HEADERS)

    assert response.status_code == 202
    _assert_single_item_response(response)

    return recid


def test_publish_draft(client, input_data):
    """Test draft publication of a non-existing record.

    It has to first create said draft.
    """
    recid = _create_and_publish(client, input_data)

    # Check draft does not exists anymore
    # FIXME: Remove import when exception is properly handled
    with pytest.raises(NoResultFound):
        response = client.get(
            "/mocks/{}/draft".format(recid), headers=HEADERS)

        assert response.status_code == 404

    # Check record exists
    response = client.get("/mocks/{}".format(recid), headers=HEADERS)

    assert response.status_code == 200

    _assert_single_item_response(response)


def test_action_not_configured(client, identity_simple):
    """Tests a non configured action call."""
    # NOTE: recid can be dummy since it won't reach pass the resource view
    response = client.post(
        "/mocks/1234-abcd/draft/actions/non-configured", headers=HEADERS
    )
    assert response.status_code == 404


def test_command_not_implemented(client, identity_simple):
    """Tests a configured action without implemented function."""
    # NOTE: recid can be dummy since it won't reach pass the resource view
    response = client.post(
        "/mocks/1234-abcd/draft/actions/command", headers=HEADERS
    )
    assert response.status_code == 500


#
# Flow tests (Note that operations are tested above
# therefore these tests do not assert their output)
#

@pytest.mark.skip()  # WHY: Draft::is_deleted is not implemented
def test_create_publish_new_revision(client, input_data,
                                     identity_simple):
    """Test draft creation of an existing record and publish it."""
    recid = _create_and_publish(client, input_data)

    # Create new draft of said record
    orig_title = input_data["metadata"]["title"]
    input_data["metadata"]["title"] = "Edited title"
    response = client.post(
        "/mocks/{}/draft".format(recid),
        headers=HEADERS
    )

    assert response.status_code == 201
    assert response.json['revision_id'] == 4
    _assert_single_item_response(response)

    # Update that new draft
    response = client.put(
        "/mocks/{}/draft".format(recid),
        data=json.dumps(input_data),
        headers=HEADERS
    )

    assert response.status_code == 200

    # Check the actual record was not modified
    response = client.get(
        "/mocks/{}".format(recid), headers=HEADERS)

    assert response.status_code == 200
    _assert_single_item_response(response)
    assert response.json['metadata']['title'] == orig_title

    # Publish it to check the increment in reversion
    response = client.post(
        "/mocks/{}/draft/actions/publish".format(recid), headers=HEADERS)

    assert response.status_code == 202
    _assert_single_item_response(response)

    assert response.json['id'] == recid
    assert response.json['revision_id'] == 2
    assert response.json['metadata']['title'] == \
        input_data["metadata"]["title"]

    # Check it was actually edited
    response = client.get(
        "/mocks/{}".format(recid), headers=HEADERS)

    assert response.json['metadata']['title'] == \
        input_data["metadata"]["title"]


@pytest.mark.skip()  # WHY: Draft::is_deleted is not implemented
def test_mutiple_edit(client, identity_simple, input_data):
    """Test the revision_id when editing record multiple times.

    This tests the `edit` service method.
    """
    # Needs `app` context because of invenio_access/permissions.py#166
    recid = _create_and_publish(client, input_data)

    # Create new draft of said record
    response = client.post(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert response.status_code == 201
    assert response.json['revision_id'] == 4

    # Request a second edit. Get the same draft (revision_id)
    response = client.post(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert response.status_code == 201
    assert response.json['revision_id'] == 4

    # Publish it to check the increment in version_id
    response = client.post(
        "/mocks/{}/draft/actions/publish".format(recid), headers=HEADERS)

    assert response.status_code == 202

    # Edit again
    response = client.post(
        "/mocks/{}/draft".format(recid), headers=HEADERS)

    assert response.status_code == 201
    assert response.json['revision_id'] == 7


def test_create_publish_new_version(client, input_data,
                                    identity_simple):
    """Creates a new version of a record.

    Publishes the draft to obtain 2 versions of a record.
    """
    recid = _create_and_publish(client, input_data)

    # Create new draft of said record
    response = client.post(
        "/mocks/{}/versions".format(recid), headers=HEADERS)

    assert response.status_code == 201
    _assert_single_item_response(response)
    assert response.json['revision_id'] == 1
    recid_2 = response.json['id']

    # Publish it to check the increment in version
    response = client.post(
        "/mocks/{}/draft/actions/publish".format(recid_2), headers=HEADERS)

    assert response.status_code == 202
    _assert_single_item_response(response)

    assert response.json['id'] == recid_2 != recid
    assert response.json['revision_id'] == 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Record schema."""

from invenio_records_resources.services.records.schema import \
    RecordSchema as RecordSchemaBase
from marshmallow import fields


class RecordSchema(RecordSchemaBase):
    """Schema for records in JSON."""

    conceptid = fields.Str()
    expires_at = fields.Str()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Draft File Resource."""

from flask_resources import CollectionResource, SingletonResource
from flask_resources.context import resource_requestctx

from ..services import FileMetadataService, FileService
from .file_config import DraftFileActionResourceConfig, DraftFileResourceConfig


class DraftFileResource(CollectionResource):
    """Draft file resource."""

    default_config = DraftFileResourceConfig

    def __init__(self, service=None, *args, **kwargs):
        """Constructor."""
        super(DraftFileResource, self).__init__(*args, **kwargs)
        self.service = service or FileMetadataService()

    # List level
    def search(self, *args, **kwargs):
        """Search over items."""
        # TODO: IMPLEMENT ME!
        return self.service.search(), 200

    def create(self, *args, **kwargs):
        """Create an item."""
        # TODO: IMPLEMENT ME!
        return self.service.create(), 200

    def update_all(self, *args, **kwargs):
        """Delete an item."""
        # TODO: IMPLEMENT ME!
        return self.service.update_all(), 200

    def delete_all(self, *args, **kwargs):
        """Delete an item."""
        # TODO: IMPLEMENT ME!
        return self.service.delete_all(), 200

    # Item level
    def read(self, *args, **kwargs):
        """Read an item."""
        # TODO: IMPLEMENT ME!
        return self.service.read(), 200

    def update(self, *args, **kwargs):
        """Update an item."""
        # TODO: IMPLEMENT ME!
        return self.service.update(), 200

    def delete(self, *args, **kwargs):
        """Delete an item."""
        # TODO: IMPLEMENT ME!
        return self.service.delete(), 200


class DraftFileActionResource(SingletonResource):
    """Draft file action resource."""

    default_config = DraftFileResourceConfig

    def __init__(self, service=None, *args, **kwargs):
        """Constructor."""
        super(DraftFileResource, self).__init__(*args, **kwargs)
        self.service = service or FileService()

    def read(self, *args, **kwargs):
        """Read an item."""
        if resource_requestctx.route["action"] == "download":
            return self.service.read()
        return {}, 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Draft File config."""


from flask_resources.resources import ResourceConfig

from ..services import FileMetadataService, FileService


class DraftFileResourceConfig(ResourceConfig):
    """Draft file resource config."""

    list_route = "/records/<pid_value>/draft/files"
    item_route = "/records/<pid_value>/draft/files/<key>"


class DraftFileActionResourceConfig(ResourceConfig):
    """Draft file action resource config."""

    # QUESTIONs:
    # 1- Should the item_route be used for SingletonResource actually? Change
    #    in Flask-Resource would be needed
    # 2- Should the list_route instead precede download with "actions" to be in
    #    keeping with other actions endpoints?
    list_route = "/records/<pid_value>/draft/files/<key>/<action>"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Draft Models API."""

from datetime import datetime

from invenio_db import db
from invenio_records.models import RecordMetadataBase
from sqlalchemy.dialects import mysql


class DraftMetadataBase(RecordMetadataBase):
    """Represent a base class for draft metadata."""

    fork_version_id = db.Column(db.Integer)
    """Version ID of the record."""

    expires_at = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=datetime.utcnow,
        nullable=True
    )
    """Specifies when the draft expires. If `NULL` the draft doesn't expire."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio Deposits Resources."""

from flask_resources import CollectionResource

from .userrecords_config import UserRecordsResourceConfig


class UserRecordsResource(CollectionResource):
    """User records resource."""

    default_config = UserRecordsResourceConfig

    def search(self, *args, **kwargs):
        """Perform a search over the items.

        GET /user/records
        """
        return STUB_LIST_RESULT



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio Drafts Resources module to create REST APIs."""

from .drafts import DraftActionResource, DraftActionResourceConfig, \
    DraftLinksSchema, DraftResource, DraftResourceConfig, \
    DraftVersionResource, DraftVersionResourceConfig
from .records import RecordResource, RecordResourceConfig
from .userrecords import UserRecordsResource
from .userrecords_config import UserRecordsResourceConfig

__all__ = (
    "DraftActionResource",
    "DraftActionResourceConfig",
    "DraftLinksSchema",
    "DraftResource",
    "DraftResourceConfig",
    "DraftVersionResource",
    "DrafVersiontResourceConfig",
    "RecordResource",
    "RecordResourceConfig",
    "UserRecordsResource",
    "UserRecordsResourceConfig",
)
